/*
 * Copyright (c) 2004-2007 The Trustees of Indiana University and Indiana
 *                         University Research and Technology
 *                         Corporation.  All rights reserved.
 * Copyright (c) 2004-2009 The University of Tennessee and The University
 *                         of Tennessee Research Foundation.  All rights
 *                         reserved.
 * Copyright (c) 2004-2005 High Performance Computing Center Stuttgart, 
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2004-2005 The Regents of the University of California.
 *                         All rights reserved.
 * $COPYRIGHT$
 * 
 * Additional copyrights may follow
 * 
 * $HEADER$
 */
/**
 * @file
 */
#ifndef MCA_BTL_HLINK_H
#define MCA_BTL_HLINK_H

#include "ompi_config.h"

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif  /* HAVE_STDLIB_H */
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif  /* HAVE_SYS_TYPES_H */
#include "opal/mca/event/event.h"
#include "ompi/class/ompi_free_list.h"
#include "ompi/mca/btl/btl.h"
#include "ompi/mca/btl/base/base.h" 
#include "ompi/mca/btl/hlink/btl_hlink_fifo.h" 
#include "mpm_transport.h"

//#define BTL_HLINK_VERBOSE

#define BTL_HLINK_MAX_PROCESS          4
#define BTL_HLINK_MAX_LOCAL_PROCESS    1
#define BTL_HLINK_EAGER_SIZE    (8*1024)
#define BTL_HLINK_MAX_SEND_SIZE (64*1024)
//These numbers are for 32-bit words
#define BTL_HLINK_MSG_OFFSET     0x00010000
#define BTL_HLINK_MMAP_SIZE      0x00400000
#define BTL_HLINK_MSMC_MMAP_SIZE 0x00010000 
#define BTL_HLINK_PAYLOAD_SIZE   0x00200000
//#define BTL_HLINK_MAX_FIFO_ITEMS (64/BTL_HLINK_MAX_LOCAL_PROCESS)
#define BTL_HLINK_MAX_FIFO_ITEMS 32

//DDR
#define BTL_HLINK_DDR_LOC_ADDRESS   0xA1000000
#define BTL_HLINK_DDR_RMT_ADDRESS   0xA1000000
#define BTL_HLINK_DDR_LOC1_ADDRESS  0xA1400000
#define BTL_HLINK_DDR_RMT1_ADDRESS  0xA1400000
#define BTL_HLINK_DDR_BRIDGE_ADDRESS  0xA1800000
//MSMC
#define BTL_HLINK_MSMC_LOC_ADDRESS    0x0c010000
#define BTL_HLINK_MSMC_RMT_ADDRESS    0x0c010000
#define BTL_HLINK_MSMC_LOC1_ADDRESS   0x0c020000
#define BTL_HLINK_MSMC_RMT1_ADDRESS   0x0c020000
#define BTL_HLINK_MSMC_BRIDGE_ADDRESS 0x0c030000
//Memory block used for EDMA transfers: user-space->(memcpy)->localMSMC->(EDMA)->(Hyplnk)->remoteDDR
#define BTL_HLINK_MSMC_EDMA_TRANSFER_ADDRESS 0x0c040000

/** Slayton has following connections:
 **	SOC1-Port0 to SOC2-Port1     
 **	SOC2-Port0 to SOC3-Port0      
 **	SOC3-Port1 to SOC4-Port1     
 **	SOC4-Port0 to SOC1-Port1     
 **
 ** Two bridges (two directions: 1->2->3 and 1<-2<-3) 
 ** (1) SOC1 communicates with SOC3 using Port0 (SOC1 writes over Port0 to BRIDGE memory in SOC2, SOC3 reads over Port0 from BRIDGE memory in SOC2)
 ** (2) SOC3 communicates with SOC1 using Port1 (SOC3 writes over Port1 to BRIDGE memory in SOC4, SOC1 reads over Port1 from BRIDGE memory in SOC4)
 ** (3) SOC2 communicates with SOC4 using Port1 (SOC2 writes over Port1 to BRIDGE memory in SOC1, SOC4 reads over Port0 from BRIDGE memory in SOC1)
 ** (4) SOC4 communicates with SOC2 using Port0 (SOC4 writes over Port1 to BRIDGE memory in SOC3, SOC2 reads over Port0 from BRIDGE memory in SOC3)
 ** SOC1, writes over Port0/reads over Port1, communicates w/ SOC3 : writes to bridge memory of SOC2; reads bridge memory of SOC4
 ** SOC2, writes over Port1/reads over Port0, communicates w/ SOC4 : writes to bridge memory of SOC1; reads bridge memory of SOC3 
 ** SOC3, writes over Port1/reads over Port0, communicates w/ SOC1 : writes to bridge memory of SOC4; reads bridge memory of SOC2
 ** SOC4, writes over Port1/reads over Port0, communicates w/ SOC2 : writes to bridge memory of SOC3; reads bridge memory of SOC1
**/

#define BTL_HLINK_MSMC_EDMA_MMAP_SIZE (BTL_HLINK_MAX_SEND_SIZE + 0x1000)

BEGIN_C_DECLS

struct mca_btl_base_endpoint_t {
    opal_list_item_t super;
    void *btl;
    opal_mutex_t lock;
    opal_list_t frag_wait_list;
    bool wait_listed;
    int peer_rank;  /**< My peer's SMP process rank.  Used for accessing SMP specfic data structures. */
    int bridging;
    int hlink_iface;
};

typedef struct mca_btl_base_endpoint_t mca_btl_base_endpoint_t;


/**
 * Shared Memory (SELF) BTL module.
 */
struct mca_btl_hlink_component_t {
    mca_btl_base_component_2_0_0_t super;  /**< base BTL component */
    int free_list_num;                     /**< initial size of free lists */
    int free_list_max;                     /**< maximum size of free lists */
    int free_list_inc;                     /**< number of elements to alloc when growing free lists */
    opal_mutex_t hlink_lock;

    ompi_free_list_t hlink_frags_eager;   /**< free list of hlink first */
    ompi_free_list_t hlink_frags_max;     /**< free list of hlink second */
    struct mca_btl_base_endpoint_t **hlink_peers;

    int node_id;                          /**< Slayton node_id is 1..4; EVM is always 1 */
    unsigned int msmc_edma_transfer_addr;
    void *hl_win_edma_transfer;
    size_t endpoint_count;
    size_t eager_limit;                /**< first fragment size */
    size_t max_frag_size;              /**< maximum (second and beyone) fragment size */

    int32_t num_smp_procs;             /**< current number of smp procs on this host */
    int32_t fd_dev_mem;

    int    num_hlink_btls;
    struct mca_btl_hlink_t      **hlink_btls;

    void  *hl_win_loc;                 /**< pointer to base pointers in DDR shared memory, visible from hyplnk0 */
    void  *hl_win_msmc_loc;            /**< pointer to base pointers in MSMC shared memory, visible from hyplnk0 */
    void  *hl_win_rmt;                 /**< pointer to base pointers in remote DDR memory, over hyplnk0 */ 
    void  *hl_win_msmc_rmt;            /**< pointer to base pointers in remote MSMC memory, over hyplnk0 */ 
    void  *hl_win_rmt_hostname;               
    void  *hl_win_loc_msgbase[BTL_HLINK_MAX_LOCAL_PROCESS];               
    void  *hl_win_rmt_msgbase[BTL_HLINK_MAX_LOCAL_PROCESS];               
    int    hl_win_rmt_msg_idx[BTL_HLINK_MAX_LOCAL_PROCESS];
    unsigned int ddr_loc_addr;
    unsigned int ddr_rmt_addr;
    unsigned int msmc_loc_addr;
    unsigned int msmc_rmt_addr;
    hlink_fifo_t *fifo[2][BTL_HLINK_MAX_LOCAL_PROCESS]; /**< cached copy of the pointer to the 2D
                                          fifo array.  The address in the shared
                                          memory segment sm_ctl_header is a relative,
                                          but this one, in process private memory, is
                                          a real virtual address */

    void  *hl_win_loc1;                /**< pointer to base pointers in DDR shared memory, visible from hyplnk1 */
    void  *hl_win_msmc_loc1;           /**< pointer to base pointers in MSMC shared memory, visible from hyplnk1 */
    void  *hl_win_rmt1;                /**< pointer to base pointers in remote DDR memory, over hyplnk1 */
    void  *hl_win_msmc_rmt1;           /**< pointer to base pointers in remote MSMC memory, over hyplnk1 */
    void  *hl_win_rmt1_hostname;               
    void  *hl_win_loc1_msgbase[BTL_HLINK_MAX_LOCAL_PROCESS];
    void  *hl_win_rmt1_msgbase[BTL_HLINK_MAX_LOCAL_PROCESS];               
    int    hl_win_rmt1_msg_idx[BTL_HLINK_MAX_LOCAL_PROCESS];
    unsigned int ddr_loc1_addr;
    unsigned int ddr_rmt1_addr;
    unsigned int msmc_loc1_addr;
    unsigned int msmc_rmt1_addr;
        
    hlink_fifo_t *fifo1[2][BTL_HLINK_MAX_LOCAL_PROCESS]; /**< cached copy of the pointer to the 2D
                                          fifo array.  The address in the shared
                                          memory segment sm_ctl_header is a relative,
                                          but this one, in process private memory, is
                                          a real virtual address */

    /*** BRIDGING (enable diagonal links on Slayton board) ***/
    unsigned int ddr_rmt_bridge_addr;        /**< area in DDR (data blobs) used for diagonal link, on transfer node (every Slayton node is transfer node) */
    unsigned int msmc_rmt_bridge_addr;       /**< area in MSMC (mails) used for diaginal link, on transfer node (every Slayton node is transfer node) */
    void  *hl_win_rmt_rd_bridge_hostname;       /**< find out diagonal peer */
    /*** Bridge to diagonal node over HLx (writting) ***/
    void  *hl_win_rmt_wr_bridge;                /**< pointer to base pointers in remote DDR memory, over hyplnk1 */
    void  *hl_win_rmt_rd_bridge;                /**< pointer to user space pointers in remote DDR memory, over hyplnk1 */
    void  *hl_win_msmc_rmt_wr_bridge;           /**< pointer to base pointers in remote MSMC memory, over hyplnk1 */
    void  *hl_win_msmc_rmt_rd_bridge;           /**< pointer to user space in remote MSMC memory, over hyplnk1 */
    int    hl_win_rmt_wr_bridge_msg_idx[BTL_HLINK_MAX_LOCAL_PROCESS];
    void  *hl_win_rmt_wr_bridge_msgbase[BTL_HLINK_MAX_LOCAL_PROCESS];
    void  *hl_win_rmt_rd_bridge_msgbase[BTL_HLINK_MAX_LOCAL_PROCESS];

    hlink_fifo_t *fifo_bridge[2][BTL_HLINK_MAX_LOCAL_PROCESS]; /**< cached copy of the pointer to the 2D
                                          fifo array.  The address in the shared
                                          memory segment sm_ctl_header is a relative,
                                          but this one, in process private memory, is
                                          a real virtual address */
    /*** BRIDGING (enable diagonal links on Slayton board) ***/

    size_t fifo_size;                    /**< number of FIFO queue entries */
    int my_rank;
    int hlink_nprocs;
    int use_hl0;
    int use_hl1;
    int use_edma;
    /* MPM-TRANSPORT interfacing */
    void *mpm_transport_shmem;
    void *mpm_transport_hyplnk0;
    void *mpm_transport_hyplnk1;
};
typedef struct mca_btl_hlink_component_t mca_btl_hlink_component_t;
OMPI_MODULE_DECLSPEC extern mca_btl_hlink_component_t mca_btl_hlink_component;


struct mca_btl_hlink_t {
    mca_btl_base_module_t super;
    void *device;
    struct mca_btl_base_endpoint_t **endpoints;

    uint32_t reg_max;
    uint32_t reg_count;
}; 
typedef struct mca_btl_hlink_t mca_btl_hlink_t;
OMPI_MODULE_DECLSPEC extern mca_btl_hlink_t mca_btl_hlink;


/**
 * Register shared memory module parameters with the MCA framework
 */
int mca_btl_hlink_component_open(void);

/**
 * Any final cleanup before being unloaded.
 */
int mca_btl_hlink_component_close(void);

/**
 * SELF module initialization.
 * 
 * @param num_btls (OUT)                  Number of BTLs returned in BTL array.
 * @param enable_progress_threads (IN)    Flag indicating whether BTL is allowed to have progress threads
 * @param enable_mpi_threads (IN)         Flag indicating whether BTL must support multilple simultaneous invocations from different threads
 *
 */
mca_btl_base_module_t** mca_btl_hlink_component_init(
    int *num_btls, 
    bool enable_progress_threads,
    bool enable_mpi_threads
);

/**
 * Cleanup any resources held by the BTL.
 * 
 * @param btl  BTL instance.
 * @return     OMPI_SUCCESS or error status on failure.
 */

int mca_btl_hlink_finalize(
    struct mca_btl_base_module_t* btl
);


/**
 * PML->BTL notification of change in the process list.
 * PML->BTL Notification that a receive fragment has been matched.
 * Called for message that is send from process with the virtual
 * address of the shared memory segment being different than that of
 * the receiver.
 * 
 * @param btl (IN)
 * @param proc (IN)  
 * @param peer (OUT)
 * @return     OMPI_SUCCESS or error status on failure.
 * 
 */

int mca_btl_hlink_add_procs(
    struct mca_btl_base_module_t* btl,
    size_t nprocs,
    struct ompi_proc_t **procs,
    struct mca_btl_base_endpoint_t** peers,
    struct opal_bitmap_t* reachability
);


/**
 * PML->BTL notification of change in the process list.
 *
 * @param btl (IN)     BTL instance
 * @param proc (IN)    Peer process
 * @param peer (IN)    Peer addressing information.
 * @return             Status indicating if cleanup was successful
 *
 */
int mca_btl_hlink_del_procs(
    struct mca_btl_base_module_t* btl,
    size_t nprocs,
    struct ompi_proc_t **procs,
    struct mca_btl_base_endpoint_t **peers
);


/**
 * Allocate a segment.
 *
 * @param btl (IN)      BTL module
 * @param size (IN)     Request segment size.
 */
mca_btl_base_descriptor_t* mca_btl_hlink_alloc(
    struct mca_btl_base_module_t* btl,
    struct mca_btl_base_endpoint_t* endpoint,
    uint8_t order,
    size_t size,
    uint32_t flags
);

/**
 * Return a segment allocated by this BTL.
 *
 * @param btl (IN)      BTL module
 * @param segment (IN)  Allocated segment.
 */
int mca_btl_hlink_free(
    struct mca_btl_base_module_t* btl,
    mca_btl_base_descriptor_t* segment
);
                                                                                                                   
/**
 * Pack data
 *
 * @param btl (IN)      BTL module
 * @param peer (IN)     BTL peer addressing
 */
struct mca_btl_base_descriptor_t* mca_btl_hlink_prepare_src(
    struct mca_btl_base_module_t* btl,
    struct mca_btl_base_endpoint_t* endpoint,
    struct mca_mpool_base_registration_t* registration,
    struct opal_convertor_t* convertor,
    uint8_t order,
    size_t reserve,
    size_t* size,
    uint32_t flags
);
                                                                                                            
/**
 * Prepare data for RDMA
 *
 * @param btl (IN)      BTL module
 * @param peer (IN)     BTL peer addressing
 */
struct mca_btl_base_descriptor_t* mca_btl_hlink_prepare_dst(
    struct mca_btl_base_module_t* btl,
    struct mca_btl_base_endpoint_t* endpoint,
    struct mca_mpool_base_registration_t* registration,
    struct opal_convertor_t* convertor,
    uint8_t order,
    size_t reserve,
    size_t* size,
    uint32_t flags
);
            

/**
 * Initiate an inlined send to the peer or return a descriptor.
 *
 * @param btl (IN)      BTL module
 * @param peer (IN)     BTL peer addressing
 */
extern int mca_btl_hlink_sendi( struct mca_btl_base_module_t* btl,
                             struct mca_btl_base_endpoint_t* endpoint,
                             struct opal_convertor_t* convertor,
                             void* header,
                             size_t header_size,
                             size_t payload_size,
                             uint8_t order,
                             uint32_t flags,
                             mca_btl_base_tag_t tag,
                             mca_btl_base_descriptor_t** descriptor );

/**
 * Initiate a send to the peer.
 *
 * @param btl (IN)      BTL module
 * @param peer (IN)     BTL peer addressing
 */
int mca_btl_hlink_send(
    struct mca_btl_base_module_t* btl,
    struct mca_btl_base_endpoint_t* endpoint,
    struct mca_btl_base_descriptor_t* descriptor,
    mca_btl_base_tag_t tag
);

/**
 * Initiate a put to the peer.
 *
 * @param btl (IN)      BTL module
 * @param peer (IN)     BTL peer addressing
 */
                                                                                                            
int mca_btl_hlink_rdma(
    struct mca_btl_base_module_t* btl,
    struct mca_btl_base_endpoint_t* endpoint,
    struct mca_btl_base_descriptor_t* descriptor
);

/***************************************************************************************************/
static inline unsigned int btl_hlink_align64(unsigned int non_aligned_idx)
{
   return ((non_aligned_idx + sizeof(uint64_t) - 1) / sizeof(uint64_t)) * sizeof(uint64_t);
}
static inline unsigned int btl_hlink_align32(unsigned int non_aligned_idx)
{
   return ((non_aligned_idx + sizeof(uint32_t) - 1) / sizeof(uint32_t)) * sizeof(uint32_t);
}
/***************************************************************************************************/

/**
 * Fault Tolerance Event Notification Function
 * @param state Checkpoint Stae
 * @return OMPI_SUCCESS or failure status
 */
int mca_btl_hlink_ft_event(int state);

END_C_DECLS

#endif

