/*
 * Copyright (c) 2004-2007 The Trustees of Indiana University and Indiana
 *                         University Research and Technology
 *                         Corporation.  All rights reserved.
 * Copyright (c) 2004-2006 The University of Tennessee and The University
 *                         of Tennessee Research Foundation.  All rights
 *                         reserved.
 * Copyright (c) 2004-2005 High Performance Computing Center Stuttgart, 
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2004-2005 The Regents of the University of California.
 *                         All rights reserved.
 * Copyright (c) 2010      Cisco Systems, Inc. All rights reserved.
 * $COPYRIGHT$
 * 
 * Additional copyrights may follow
 * 
 * $HEADER$
 */
#include "ompi_config.h"
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif  /* HAVE_UNISTD_H */
#include <stddef.h>
#ifdef HAVE_STRING_H
#include <string.h>
#endif  /* HAVE_STRING_H */
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif  /* HAVE_SYS_TYPES_H */
#include <sys/mman.h>
#include <dlfcn.h>

#include "ompi/constants.h"
#include "opal/runtime/opal.h"
#include "opal/mca/event/event.h"
#include "opal/mca/base/mca_base_param.h"
#include "btl_hlink.h"
#include "btl_hlink_frag.h"

static int mca_btl_hlink_component_progress (void);
static int hlink_register(void);
/***********************************************************************/
static opal_event_t btl_hlink_term_handler1;
static opal_event_t btl_hlink_term_handler2;
static void btl_hlink_shutdown_callback(int fd, short flags, void *arg)
{
   opal_output(0, "BTL_HLINK_SHUTDOWN_CALLBACK!!!");
}
/***********************************************************************/
/*
 * Shared Memory (SELF) component instance. 
 */

mca_btl_hlink_component_t mca_btl_hlink_component = {
    {  /* super is being filled in */
        /* First, the mca_base_component_t struct containing meta information
          about the component itself */
        {
            MCA_BTL_BASE_VERSION_2_0_0,

            "hlink", /* MCA component name */
            OMPI_MAJOR_VERSION,  /* MCA component major version */
            OMPI_MINOR_VERSION,  /* MCA component minor version */
            OMPI_RELEASE_VERSION,  /* MCA component release version */
            mca_btl_hlink_component_open,  /* component open */
            mca_btl_hlink_component_close,  /* component close */
            NULL,
            hlink_register,
        },
        {
            /* The component is checkpoint ready */
            MCA_BASE_METADATA_PARAM_CHECKPOINT
        },

        mca_btl_hlink_component_init, 
        mca_btl_hlink_component_progress,
    }  /* end super */
};
/**************************************************************************************/
static inline char* mca_btl_hlink_param_register_string(
    const char* param_name,
    const char* default_value)
{
    char *param_value;
    int id = mca_base_param_register_string("btl","hlink",param_name,NULL,default_value);
    mca_base_param_lookup_string(id, &param_value);
    return param_value;
}

static inline int mca_btl_hlink_param_register_int(
    const char* param_name,
    int default_value)
{
    int id = mca_base_param_register_int("btl","hlink",param_name,NULL,default_value);
    int param_value = default_value;
    mca_base_param_lookup_int(id,&param_value);
    return param_value;
}

static int hlink_register(void)
{
    int i;

    opal_output_verbose(20, mca_btl_base_output, "BTL_HLINK TIMPIDBG: hlink_component_register!!!");
    /**************************************/
    mca_btl_hlink_component.node_id = '1';
    { /** Find location in Slayton topology: if this is EVM, default to 1 **/
      FILE *node_id_detect = fopen("/proc/device-tree/sl1500/node/id", "rb");
      unsigned char node_id_buff[2];
      if(node_id_detect)
      {
         fread(node_id_buff, 1, 2, node_id_detect);
         mca_btl_hlink_component.node_id = node_id_buff[0];
         fclose (node_id_detect);
      } else {
        opal_output_verbose(100, mca_btl_base_output, "This is EVM, using hl0 only!");
      }
    }

    /* register HLINK component parameters */
    if(mca_btl_hlink_component.node_id == '1')
    {
    mca_btl_hlink_component.ddr_loc_addr =
        mca_btl_hlink_param_register_int("ddr_loc_addr", BTL_HLINK_DDR_LOC_ADDRESS);
    mca_btl_hlink_component.msmc_loc_addr =
        mca_btl_hlink_param_register_int("msmc_loc_addr", BTL_HLINK_MSMC_LOC_ADDRESS);
    mca_btl_hlink_component.ddr_rmt_addr =
        mca_btl_hlink_param_register_int("ddr_rmt_addr", BTL_HLINK_DDR_RMT_ADDRESS);
    mca_btl_hlink_component.msmc_rmt_addr =
        mca_btl_hlink_param_register_int("msmc_rmt_addr", BTL_HLINK_MSMC_RMT_ADDRESS);

    mca_btl_hlink_component.ddr_loc1_addr =
        mca_btl_hlink_param_register_int("ddr_loc1_addr", BTL_HLINK_DDR_LOC1_ADDRESS);
    mca_btl_hlink_component.msmc_loc1_addr =
        mca_btl_hlink_param_register_int("msmc_loc1_addr", BTL_HLINK_MSMC_LOC1_ADDRESS);
    mca_btl_hlink_component.ddr_rmt1_addr =
        mca_btl_hlink_param_register_int("ddr_rmt1_addr", BTL_HLINK_DDR_RMT1_ADDRESS);
    mca_btl_hlink_component.msmc_rmt1_addr =
        mca_btl_hlink_param_register_int("msmc_rmt1_addr", BTL_HLINK_MSMC_RMT1_ADDRESS);
    } else { /** Mirrored addresses **/
    mca_btl_hlink_component.ddr_loc1_addr =
        mca_btl_hlink_param_register_int("ddr_loc_addr", BTL_HLINK_DDR_LOC_ADDRESS);
    mca_btl_hlink_component.msmc_loc1_addr =
        mca_btl_hlink_param_register_int("msmc_loc_addr", BTL_HLINK_MSMC_LOC_ADDRESS);
    mca_btl_hlink_component.ddr_rmt1_addr =
        mca_btl_hlink_param_register_int("ddr_rmt_addr", BTL_HLINK_DDR_RMT_ADDRESS);
    mca_btl_hlink_component.msmc_rmt1_addr =
        mca_btl_hlink_param_register_int("msmc_rmt_addr", BTL_HLINK_MSMC_RMT_ADDRESS);

    mca_btl_hlink_component.ddr_loc_addr =
        mca_btl_hlink_param_register_int("ddr_loc1_addr", BTL_HLINK_DDR_LOC1_ADDRESS);
    mca_btl_hlink_component.msmc_loc_addr =
        mca_btl_hlink_param_register_int("msmc_loc1_addr", BTL_HLINK_MSMC_LOC1_ADDRESS);
    mca_btl_hlink_component.ddr_rmt_addr =
        mca_btl_hlink_param_register_int("ddr_rmt1_addr", BTL_HLINK_DDR_RMT1_ADDRESS);
    mca_btl_hlink_component.msmc_rmt_addr =
        mca_btl_hlink_param_register_int("msmc_rmt1_addr", BTL_HLINK_MSMC_RMT1_ADDRESS);
    }

    mca_btl_hlink_component.msmc_edma_transfer_addr =
        mca_btl_hlink_param_register_int("msmc_edma_transfer_addr", BTL_HLINK_MSMC_EDMA_TRANSFER_ADDRESS);
    mca_btl_hlink_component.msmc_rmt_bridge_addr =
        mca_btl_hlink_param_register_int("msmc_rmt_bridge_addr", BTL_HLINK_MSMC_BRIDGE_ADDRESS);
    mca_btl_hlink_component.ddr_rmt_bridge_addr =
        mca_btl_hlink_param_register_int("ddr_rmt_bridge_addr", BTL_HLINK_DDR_BRIDGE_ADDRESS);

    mca_btl_hlink_component.fifo_size =
        mca_btl_hlink_param_register_int("fifo_size", BTL_HLINK_MAX_FIFO_ITEMS);
    mca_btl_hlink_component.use_hl0  = 
        mca_btl_hlink_param_register_int("use_hl0", 1);
    mca_btl_hlink_component.use_hl1  = 
        mca_btl_hlink_param_register_int("use_hl1", 1);
    mca_btl_hlink_component.use_edma  = 
        mca_btl_hlink_param_register_int("use_edma", 0);  // Default: use edma

//    mca_btl_sm_component.sm_mpool_name =
//        mca_btl_sm_param_register_string("mpool", "sm");
//    mca_base_param_reg_int( (mca_base_component_t*)&mca_btl_hlink_component, "free_list_num",
//                            "Number of fragments by default", false, false,
//                            0, &mca_btl_hlink_component.free_list_num );

    /* register SELF component parameters */
    mca_base_param_reg_int( (mca_base_component_t*)&mca_btl_hlink_component, "free_list_num",
                            "Number of fragments by default", false, false,
                            0, &mca_btl_hlink_component.free_list_num );
    mca_base_param_reg_int( (mca_base_component_t*)&mca_btl_hlink_component, "free_list_max",
                            "Maximum number of fragments", false, false,
                            -1, &mca_btl_hlink_component.free_list_max );
    mca_base_param_reg_int( (mca_base_component_t*)&mca_btl_hlink_component, "free_list_inc",
                            "Increment by this number of fragments", false, false,
                            32, &mca_btl_hlink_component.free_list_inc );

      return OMPI_SUCCESS;
}
/*
 *  Called by MCA framework to open the component, registers
 *  component parameters.
 */
int mca_btl_hlink_component_open(void)
{
    opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK TIMPIDBG: hlink_component_open!!!");
    mca_btl_hlink.super.btl_exclusivity = MCA_BTL_EXCLUSIVITY_HIGH;
    mca_btl_hlink.super.btl_eager_limit = BTL_HLINK_EAGER_SIZE;
    mca_btl_hlink.super.btl_rndv_eager_limit = BTL_HLINK_EAGER_SIZE;
    mca_btl_hlink.super.btl_max_send_size = BTL_HLINK_MAX_SEND_SIZE;
    mca_btl_hlink.super.btl_flags = MCA_BTL_FLAGS_SEND | MCA_BTL_FLAGS_SEND_INPLACE;
    mca_btl_hlink.super.btl_seg_size = sizeof (mca_btl_base_segment_t);
    mca_btl_hlink.super.btl_bandwidth = 40000;
    mca_btl_hlink.super.btl_latency   = 1;
    mca_btl_base_param_register(&mca_btl_hlink_component.super.btl_version, &mca_btl_hlink);
    /**************************************/
    mca_btl_hlink_component.max_frag_size = mca_btl_hlink.super.btl_max_send_size;
    mca_btl_hlink_component.eager_limit = mca_btl_hlink.super.btl_eager_limit;
    /**************************************/
    /* initialize objects */
    OBJ_CONSTRUCT(&mca_btl_hlink_component.hlink_lock, opal_mutex_t);
    OBJ_CONSTRUCT(&mca_btl_hlink_component.hlink_frags_eager, ompi_free_list_t);
    OBJ_CONSTRUCT(&mca_btl_hlink_component.hlink_frags_max, ompi_free_list_t);
    {
       extern void *hyplnkInitCfg[];
       int ii;
#ifdef BTL_HLINK_VERBOSE
       void *lib_handle, **struct_handle;
       char *error;
       lib_handle = dlopen("/usr/lib/libhyplnk_k2h.so", RTLD_LAZY);
       if (!lib_handle) {
         opal_output(0, "BTL HLINK DLOPEN HANDLE error:%s\n", dlerror());
       } else {
         opal_output(100, "BTL HLINK DLOPEN HANDLE fine!\n");
       }
       struct_handle = dlsym(lib_handle, "hyplnkInitCfg");
       if ((error = dlerror()) != NULL)  
       {
         opal_output(0, "BTL HLINK DLSYM ERROR:%s\n", dlerror());
       } else {
         opal_output(100, "BTL HLINK DLSYM fine (%p)!\n", struct_handle);
         for (ii = 0; ii < 8; ii ++) opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK dl[%p]", struct_handle[ii]);
       }
       dlclose(lib_handle);
#endif
       opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK  BTL HLINK start of HYPLNKINITCFG: %p ", hyplnkInitCfg);
       if(hyplnkInitCfg) {
         for (ii = 0; ii < 8; ii ++) opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK [%p]", hyplnkInitCfg[ii]);
       }
       opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK BTL HLINK end of HYPLNKINITCFG\n");     
    }  
    return OMPI_SUCCESS;
}

/*
 * component cleanup - sanity checking of queue lengths
 */

int mca_btl_hlink_component_close(void)
{
    opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK TIMPIDBG: hlink_component_close!!!");
    OBJ_DESTRUCT(&mca_btl_hlink_component.hlink_lock);
    OBJ_DESTRUCT(&mca_btl_hlink_component.hlink_frags_eager);
    OBJ_DESTRUCT(&mca_btl_hlink_component.hlink_frags_max);
    return OMPI_SUCCESS;
}


/*
 *  HLINK component initialization
 */
mca_btl_base_module_t** mca_btl_hlink_component_init( int *num_btls, 
                                                     bool enable_progress_threads,
                                                     bool enable_mpi_threads )
{
    mca_btl_base_module_t **btls = NULL;
    char host_name[80], *ptr_hostname;
    int rc;
    /** mpm_transport init **/
    mpm_transport_open_t ocfg = {
      .open_mode      = (O_SYNC|O_RDWR),
      .msec_timeout   = 2000
    };

    mpm_transport_mmap_t mcfg = {
      .mmap_prot      = (PROT_READ|PROT_WRITE),
      .mmap_flags     = MAP_SHARED,
    };

    gethostname(host_name, 80); 
    opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK TIMPIDBG: hlink_component_init!!!");
    /* allocate the Shared Memory PTL */
    *num_btls = 1;
    btls = (mca_btl_base_module_t**)malloc((*num_btls)*sizeof(mca_btl_base_module_t*));
    if (NULL == btls) {
        return NULL;
    }
    mca_btl_hlink_component.num_hlink_btls = 1;
    mca_btl_hlink_component.hlink_btls     = btls;
    mca_btl_hlink_component.mpm_transport_hyplnk0 = NULL;
    mca_btl_hlink_component.mpm_transport_hyplnk1 = NULL;
    /* initialize free lists */
    rc = ompi_free_list_init_ex_new (&mca_btl_hlink_component.hlink_frags_eager,
                                     sizeof(mca_btl_hlink_frag1_t),
                                     opal_cache_line_size, OBJ_CLASS(mca_btl_hlink_frag1_t),
                                     mca_btl_hlink.super.btl_eager_limit,
                                     opal_cache_line_size,
                                     mca_btl_hlink_component.free_list_num,
                                     mca_btl_hlink_component.free_list_max,
                                     mca_btl_hlink_component.free_list_inc,
                                     NULL, (ompi_free_list_item_init_fn_t) mca_btl_hlink_frag_init,
                                     (void *) &mca_btl_hlink);

    if (OPAL_UNLIKELY(OMPI_SUCCESS != rc)) {
        opal_output(0, "error creating hlink fragment1 free list");
        return rc;
    }


    rc = ompi_free_list_init_ex_new (&mca_btl_hlink_component.hlink_frags_max,
                                     sizeof(mca_btl_hlink_frag1_t),
                                     opal_cache_line_size, OBJ_CLASS(mca_btl_hlink_frag1_t),
                                     mca_btl_hlink.super.btl_max_send_size,
                                     opal_cache_line_size,
                                     mca_btl_hlink_component.free_list_num,
                                     mca_btl_hlink_component.free_list_max,
                                     mca_btl_hlink_component.free_list_inc,
                                     NULL, (ompi_free_list_item_init_fn_t) mca_btl_hlink_frag_init,
                                     (void *) &mca_btl_hlink);

    if (OPAL_UNLIKELY(OMPI_SUCCESS != rc)) {
        opal_output(0, "error creating hlink fragment2 free list");
        return rc;
    }

    /* get pointer to the btls */
    btls[0] = (mca_btl_base_module_t *)(&mca_btl_hlink);

      /************************************************************************************************************************/
      /** SHARED MEMORY ACCESS: one MSMC region is mapped for both interfaces, and two DDR regions, one per HYPLNK interface **/
      /************************************************************************************************************************/

      /** Open local shared memory interface **/
      mca_btl_hlink_component.mpm_transport_shmem = (void *)mpm_transport_open("arm-local-shmem", &ocfg);
      if (!mca_btl_hlink_component.mpm_transport_shmem) {
        opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK shmem open failed");
        return NULL;
      }
      opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK shmem open successfull!!");
      /****-----------------------------------------------------****/
      /** Map local msmc region to user space, used over hyplnk0 **/
      /****-----------------------------------------------------****/
      mca_btl_hlink_component.hl_win_msmc_loc = mpm_transport_mmap((mpm_transport_h)mca_btl_hlink_component.mpm_transport_shmem, 
                                                                    mca_btl_hlink_component.msmc_loc_addr, BTL_HLINK_MSMC_MMAP_SIZE, &mcfg);
      if(!mca_btl_hlink_component.hl_win_msmc_loc)
      {
        opal_output_verbose(0, mca_btl_base_output, "BTL_HLINK mmap_failed_hl_win_msmc_loc (local MSMC over hyplnk0)!");
        return NULL;
      }
      /* Identify yourself to remote endpoint, over hyplnk0 - needed for add_procs */
      ptr_hostname = (char *)mca_btl_hlink_component.hl_win_msmc_loc + BTL_HLINK_MSG_OFFSET - 80;
      memset (ptr_hostname, 0x77, 80);
      strcpy (ptr_hostname, host_name);

      opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK shmem MSMC0 mmap successfull!!");
      /** Map local DDR region to user space, used over hyplnk0 **/
      mca_btl_hlink_component.hl_win_loc = mpm_transport_mmap((mpm_transport_h)mca_btl_hlink_component.mpm_transport_shmem,
                                                                    mca_btl_hlink_component.ddr_loc_addr, BTL_HLINK_MMAP_SIZE, &mcfg);
      if(!mca_btl_hlink_component.hl_win_loc)
      {
        opal_output (0, "mmap_failed_hl_win_loc (local DDR over hyplnk0)!");
        return NULL;
      }
      opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK shmem DDR0 mmap successfull!!");
      /** Map local msmc region to user space, used over hyplnk1 **/
      mca_btl_hlink_component.hl_win_msmc_loc1 = mpm_transport_mmap((mpm_transport_h)mca_btl_hlink_component.mpm_transport_shmem,
                                                                    mca_btl_hlink_component.msmc_loc1_addr, BTL_HLINK_MSMC_MMAP_SIZE, &mcfg);
      if(!mca_btl_hlink_component.hl_win_msmc_loc1)
      {
        opal_output_verbose(20, mca_btl_base_output, "BTL_HLINK mmap_failed_hl_win_msmc_loc (local MSMC, over hyplnk1)!");
        return NULL;
      }
      /* Identify yourself to remote endpoint, over hyplnk0 - needed for add_procs */
      ptr_hostname = (char *)mca_btl_hlink_component.hl_win_msmc_loc1 + BTL_HLINK_MSG_OFFSET - 80;
      memset (ptr_hostname, 0x77, 80);
      strcpy (ptr_hostname, host_name);

      opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK shmem MSMC1 mmap successfull!!");
      /** Map local DDR region to user space, used over hyplnk1 **/
      mca_btl_hlink_component.hl_win_loc1 = mpm_transport_mmap((mpm_transport_h)mca_btl_hlink_component.mpm_transport_shmem,
                                                                    mca_btl_hlink_component.ddr_loc1_addr, BTL_HLINK_MMAP_SIZE, &mcfg);

      if(!mca_btl_hlink_component.hl_win_loc1)
      {
        opal_output_verbose(20, mca_btl_base_output, "BTL_HLINK mmap_failed_hl_win_loc (local DDR, over hyplnk1)!");
        return NULL;
      }
      opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK shmem DDR1 mmap successfull!!");

      if(mca_btl_hlink_component.use_edma)
      {
        /** Map local MSMC region to user space, used for EDMA driven transfers **/
        mca_btl_hlink_component.hl_win_edma_transfer = mpm_transport_mmap((mpm_transport_h)mca_btl_hlink_component.mpm_transport_shmem,
                                                                          mca_btl_hlink_component.msmc_edma_transfer_addr, BTL_HLINK_MSMC_EDMA_MMAP_SIZE, &mcfg);

        if(!mca_btl_hlink_component.hl_win_edma_transfer)
        {
          opal_output(0, "mmap_failed_hl_win_edma_transfer!");
          return NULL;
        }
        opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK shmem MSMC of EDMA transfer mmap successfull!!");
      }
   /*********************************************************************************************/
   /** Probe and Open hyplnk0 interface                                                        **/
   /*********************************************************************************************/
   if(mca_btl_hlink_component.node_id == '1') 
   { /* This block is also executed by EVMs */
      opal_output_verbose(20, mca_btl_base_output, "BTL_HLINK attempt HyperLink0 then HyperLink1");
      if(mca_btl_hlink_component.use_hl0)
      { /* If user requested not to use Hyperlink0, by resetting use_hl0 to 0, this is skipped */
        /** Open hyplnk0 interface **/
        opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK hyplnk0 attempt opening");
        if(mca_btl_hlink_component.use_edma) {
          mca_btl_hlink_component.mpm_transport_hyplnk0 = (void *)mpm_transport_open("arm-remote-hyplnk-0", &ocfg);
        } else {
          mca_btl_hlink_component.mpm_transport_hyplnk0 = (void *)mpm_transport_open("arm-remote-hyplnk-0-no-dma", &ocfg);
        }
        if (!mca_btl_hlink_component.mpm_transport_hyplnk0) {
          opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK hyplnk0 open failed");
          mca_btl_hlink_component.use_hl0 = 0;
        } else {
          opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK hyplnk0 open successfull!!");
        }
      }
      if(mca_btl_hlink_component.use_hl1)
      { /* If user requested not to use Hyperlink1, by resetting use_hl1 to 0, this is skipped */
        /** Open hyplnk1 interface **/
        opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK hyplnk1 attempt opening");
        if(mca_btl_hlink_component.use_edma) {
          mca_btl_hlink_component.mpm_transport_hyplnk1 = (void *)mpm_transport_open("arm-remote-hyplnk-1", &ocfg);
        } else {
          mca_btl_hlink_component.mpm_transport_hyplnk1 = (void *)mpm_transport_open("arm-remote-hyplnk-1-no-dma", &ocfg);
        }
        if (!mca_btl_hlink_component.mpm_transport_hyplnk1) {
          opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK hyplnk1 open failed");
          mca_btl_hlink_component.use_hl1 = 0;
        } else {
          opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK hyplnk1 open successfull!!");
        }
      }
   } else { /* other Slayton nodes; this code is not used for EVM */
      opal_output_verbose(20, mca_btl_base_output, "BTL_HLINK attempt HyperLink1 then HyperLink0");
      if(mca_btl_hlink_component.use_hl1)
      { /* If user requested not to use Hyperlink1, by resetting use_hl1 to 0, this is skipped */
        /** Open hyplnk1 interface **/
        opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK hyplnk1 attempt opening");
        if(mca_btl_hlink_component.use_edma) {
          mca_btl_hlink_component.mpm_transport_hyplnk1 = (void *)mpm_transport_open("arm-remote-hyplnk-1", &ocfg);
        } else {
          mca_btl_hlink_component.mpm_transport_hyplnk1 = (void *)mpm_transport_open("arm-remote-hyplnk-1-no-dma", &ocfg);
        }
        if (!mca_btl_hlink_component.mpm_transport_hyplnk1) {
          opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK hyplnk1 open failed");
          mca_btl_hlink_component.use_hl1 = 0;
        } else {
          opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK hyplnk1 open successfull!!");
        }
      }

      if(mca_btl_hlink_component.use_hl0)
      { /* If user requested not to use Hyperlink0, by resetting use_hl0 to 0, this is skipped */
        /** Open hyplnk0 interface **/
        opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK hyplnk0 attempt opening\n");
        if(mca_btl_hlink_component.use_edma) {
          mca_btl_hlink_component.mpm_transport_hyplnk0 = (void *)mpm_transport_open("arm-remote-hyplnk-0", &ocfg);
        } else {
          mca_btl_hlink_component.mpm_transport_hyplnk0 = (void *)mpm_transport_open("arm-remote-hyplnk-0-no-dma", &ocfg);
        }
        if (!mca_btl_hlink_component.mpm_transport_hyplnk0) {
          opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK hyplnk0 open failed\n");
          mca_btl_hlink_component.use_hl0 = 0;
        } else {
          opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK hyplnk0 open successfull!!");
        }
      }
   }

   /*********************************************************************************************/
   opal_output_verbose(20, mca_btl_base_output, "BTL_HLINK hyplnk0=%p hyplnk1=%p\n", 
               mca_btl_hlink_component.mpm_transport_hyplnk0, mca_btl_hlink_component.mpm_transport_hyplnk1);
      mca_btl_hlink_component.hl_win_rmt_rd_bridge_hostname = NULL;
      mca_btl_hlink_component.hl_win_rmt_rd_bridge          = NULL;
      mca_btl_hlink_component.hl_win_rmt_wr_bridge          = NULL;
      /*********************************************************************************************/
      /** HYPLNK0 ACCESS: one remote MSMC region and one remote DDR regions is mapped via hyplnk0 **/
      /*********************************************************************************************/
      if(mca_btl_hlink_component.use_hl0)
      {
        /** Map remote MSMC region visible via hyplnk0 to user space **/
        mca_btl_hlink_component.hl_win_msmc_rmt = mpm_transport_mmap((mpm_transport_h)mca_btl_hlink_component.mpm_transport_hyplnk0, 
                                                                      mca_btl_hlink_component.msmc_rmt_addr, BTL_HLINK_MSMC_MMAP_SIZE, &mcfg);
        if(!mca_btl_hlink_component.hl_win_msmc_rmt)
        {
          opal_output(0, "mmap_failed_hl_win_msmc_rmt (MSMC over hyplnk0)!");
          return NULL;
        }
        opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK hyplnk0 MSMC mmap successfull!!");
        /** Map remote DDR region visible via hyplnk0 to user space **/
        mca_btl_hlink_component.hl_win_rmt = mpm_transport_mmap((mpm_transport_h)mca_btl_hlink_component.mpm_transport_hyplnk0,
                                                                 mca_btl_hlink_component.ddr_rmt_addr, BTL_HLINK_MMAP_SIZE, &mcfg);
        if(!mca_btl_hlink_component.hl_win_rmt)
        {
          opal_output (0, "mmap_failed_hl_win_rmt (DDR over hyplnk0)!");
          return NULL;
        }
        opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK hyplnk0 DDR mmap successfull!!");
        /* This is where I am going to read remote hostname visible over hyplnk0 */
        mca_btl_hlink_component.hl_win_rmt_hostname = (void *)((char *)mca_btl_hlink_component.hl_win_msmc_rmt + BTL_HLINK_MSG_OFFSET - 80);

      if(mca_btl_hlink_component.use_hl0 * mca_btl_hlink_component.use_hl1)
      {
        /*** DIAGONAL connection ***/
        /** Map remote MSMC region visible via hyplnk0 to user space **/
        if(mca_btl_hlink_component.node_id == '1')
        {  /** Node '1' does writting over Hyperlink0 **/
           mca_btl_hlink_component.hl_win_msmc_rmt_wr_bridge = mpm_transport_mmap((mpm_transport_h)mca_btl_hlink_component.mpm_transport_hyplnk0,
                                                                        mca_btl_hlink_component.msmc_rmt_bridge_addr, BTL_HLINK_MSMC_MMAP_SIZE, &mcfg);
           if(!mca_btl_hlink_component.hl_win_msmc_rmt_wr_bridge)
           {
             opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK mmap_failed_hl_win_msmc_rmt_bridge (MSMC over hyplnk0)!");
             return NULL;
           }
           opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK hyplnk0 MSMC bridge wr mmap successfull!!");
          /** Map remote DDR region visible via hyplnk0 to user space **/
          mca_btl_hlink_component.hl_win_rmt_wr_bridge = mpm_transport_mmap((mpm_transport_h)mca_btl_hlink_component.mpm_transport_hyplnk0,
                                                                 mca_btl_hlink_component.ddr_rmt_bridge_addr, BTL_HLINK_MMAP_SIZE, &mcfg);
          if(!mca_btl_hlink_component.hl_win_rmt_wr_bridge)
          {
            opal_output (0, "mmap_failed_hl_win_rmt_wr (DDR over hyplnk0)!");
            return NULL;
          }
          opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK hyplnk0 DDR mmap wr successfull!!");
          /* This is where I am indicating identity accross diagonal connection */
          /* For diagonal connections, writting is always done over hyperlink0  */
          ptr_hostname = 
            (void *)((char *)mca_btl_hlink_component.hl_win_msmc_rmt_wr_bridge + BTL_HLINK_MSG_OFFSET - 80);
          memset (ptr_hostname, 0x78, 80);
          strcpy (ptr_hostname, host_name);
        } else { /* Nodes '2','3' and '4' do reading over HL0 */
           mca_btl_hlink_component.hl_win_msmc_rmt_rd_bridge = mpm_transport_mmap((mpm_transport_h)mca_btl_hlink_component.mpm_transport_hyplnk0,
                                                                        mca_btl_hlink_component.msmc_rmt_bridge_addr, BTL_HLINK_MSMC_MMAP_SIZE, &mcfg);
           if(!mca_btl_hlink_component.hl_win_msmc_rmt_rd_bridge)
           {
             opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK mmap_failed_hl_win_msmc_rmt_rd_bridge (MSMC over hyplnk0)!");
             return NULL;
           }
           opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK hyplnk0 MSMC bridge rd mmap successfull!!");
          /** Map remote DDR region visible via hyplnk0 to user space **/
          mca_btl_hlink_component.hl_win_rmt_rd_bridge = mpm_transport_mmap((mpm_transport_h)mca_btl_hlink_component.mpm_transport_hyplnk0,
                                                                 mca_btl_hlink_component.ddr_rmt_bridge_addr, BTL_HLINK_MMAP_SIZE, &mcfg);
          if(!mca_btl_hlink_component.hl_win_rmt_rd_bridge)
          {
            opal_output (0, "mmap_failed_hl_win_rmt_rd_bridge (DDR over hyplnk0)!");
            return NULL;
          }
          opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK hyplnk0 DDR mmap successfull!!");
          mca_btl_hlink_component.hl_win_rmt_rd_bridge_hostname =
            (void *)((char *)mca_btl_hlink_component.hl_win_msmc_rmt_rd_bridge + BTL_HLINK_MSG_OFFSET - 80);
        }
      }
      }
      /*********************************************************************************************/
      /** HYPLNK1 ACCESS: one remote MSMC region and one remote DDR regions is mapped via hyplnk1 **/
      /*********************************************************************************************/
      if(mca_btl_hlink_component.use_hl1)
      {
        /** Map remote MSMC region visible via hyplnk1 to user space **/
        mca_btl_hlink_component.hl_win_msmc_rmt1 = mpm_transport_mmap((mpm_transport_h)mca_btl_hlink_component.mpm_transport_hyplnk1, 
                                                                      mca_btl_hlink_component.msmc_rmt1_addr, BTL_HLINK_MSMC_MMAP_SIZE, &mcfg);
        if(!mca_btl_hlink_component.hl_win_msmc_rmt1)
        {
          opal_output(0, "mmap_failed_hl_win_msmc_rmt (MSMC over hyplnk1)!");
          return NULL;
        }
        opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK hyplnk1 MSMC mmap successfull!!");
        /** Map remote DDR region visible via hyplnk1 to user space **/
        mca_btl_hlink_component.hl_win_rmt1 = mpm_transport_mmap((mpm_transport_h)mca_btl_hlink_component.mpm_transport_hyplnk1,
                                                                  mca_btl_hlink_component.ddr_rmt1_addr, BTL_HLINK_MMAP_SIZE, &mcfg);
        if(!mca_btl_hlink_component.hl_win_rmt1)
        {
          opal_output(0, "mmap_failed_hl_win_rmt (DDR over hyplnk1)!\n");
          return NULL;
        }
        opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK hyplnk1 DDR mmap successfull!!");
        /* This is where I am going to read remote hostname visible over hyplnk1 */
        mca_btl_hlink_component.hl_win_rmt1_hostname = (void *)((char *)mca_btl_hlink_component.hl_win_msmc_rmt1 + BTL_HLINK_MSG_OFFSET - 80);

      if(mca_btl_hlink_component.use_hl0 * mca_btl_hlink_component.use_hl1)
      {
        /*** DIAGONAL connection over Hyperlink1 ***/
        if(mca_btl_hlink_component.node_id == '1')
        {  /** Node '1' does reading over Hyperlink1 **/
           mca_btl_hlink_component.hl_win_msmc_rmt_rd_bridge = mpm_transport_mmap((mpm_transport_h)mca_btl_hlink_component.mpm_transport_hyplnk1,
                                                                        mca_btl_hlink_component.msmc_rmt_bridge_addr, BTL_HLINK_MSMC_MMAP_SIZE, &mcfg);
           if(!mca_btl_hlink_component.hl_win_msmc_rmt_rd_bridge)
           {
             opal_output(0, "mmap_failed_hl_win_msmc_rmt_rd_bridge (MSMC over hyplnk1)!");
             return NULL;
           }
           opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK hyplnk1 MSMC bridge rd mmap successfull!!");
          /** Map remote DDR region visible via hyplnk1 to user space **/
          mca_btl_hlink_component.hl_win_rmt_rd_bridge = mpm_transport_mmap((mpm_transport_h)mca_btl_hlink_component.mpm_transport_hyplnk1,
                                                                 mca_btl_hlink_component.ddr_rmt_bridge_addr, BTL_HLINK_MMAP_SIZE, &mcfg);
          if(!mca_btl_hlink_component.hl_win_rmt_rd_bridge)
          {
            opal_output (0, "mmap_failed_hl_win_rmt_rd (DDR over hyplnk1)!");
            return NULL;
          }
          opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK hyplnk1 DDR mmap rd successfull!!");
          /* This is where I am indicating identity accross diagonal connection */
          /* For diagonal connections, writting is always done over hyperlink0  */
          mca_btl_hlink_component.hl_win_rmt_rd_bridge_hostname =
            (void *)((char *)mca_btl_hlink_component.hl_win_msmc_rmt_rd_bridge + BTL_HLINK_MSG_OFFSET - 80);
        } else { /* Nodes '2','3' and '4' do writting over HL1 */
           mca_btl_hlink_component.hl_win_msmc_rmt_wr_bridge = mpm_transport_mmap((mpm_transport_h)mca_btl_hlink_component.mpm_transport_hyplnk1,
                                                                        mca_btl_hlink_component.msmc_rmt_bridge_addr, BTL_HLINK_MSMC_MMAP_SIZE, &mcfg);
           if(!mca_btl_hlink_component.hl_win_msmc_rmt_wr_bridge)
           {
             opal_output(0, "mmap_failed_hl_win_msmc_rmt_wr_bridge (MSMC over hyplnk1)!");
             return NULL;
           }
           opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK hyplnk1 MSMC bridge wr mmap successfull!!");
          /** Map remote DDR region visible via hyplnk0 to user space **/
          mca_btl_hlink_component.hl_win_rmt_wr_bridge = mpm_transport_mmap((mpm_transport_h)mca_btl_hlink_component.mpm_transport_hyplnk1,
                                                                 mca_btl_hlink_component.ddr_rmt_bridge_addr, BTL_HLINK_MMAP_SIZE, &mcfg);
          if(!mca_btl_hlink_component.hl_win_rmt_wr_bridge)
          {
            opal_output (0, "mmap_failed_hl_win_rmt_wr_bridge (DDR over hyplnk1)!");
            return NULL;
          }
          opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK hyplnk1 DDR mmap wr successfull!!");
          ptr_hostname =
            (void *)((char *)mca_btl_hlink_component.hl_win_msmc_rmt_wr_bridge + BTL_HLINK_MSG_OFFSET - 80);
          memset (ptr_hostname, 0x78, 80);
          strcpy (ptr_hostname, host_name);
        }
      }
      }

    if(!mca_btl_hlink_component.use_hl0 && !mca_btl_hlink_component.use_hl1)
    { /* If both interfaces are turned off, do not use hlink */
       opal_output(0, "HLINK turned off !!!");
       return NULL;
    }

    /* Set signal handlers to catch kill signals so we can properly clean up after ourselves. */
    //opal_event_set(opal_event_base, &btl_hlink_term_handler2, SIGINT, OPAL_EV_SIGNAL|OPAL_EV_PERSIST, btl_hlink_shutdown_callback, &btl_hlink_term_handler2);
    //opal_event_add(&btl_hlink_term_handler2, NULL);

    opal_output_verbose(20, mca_btl_base_output, "BTL_HLINK BTL HLINK turned ON (hl0=%d hl1=%d)!!!", 
      mca_btl_hlink_component.use_hl0, mca_btl_hlink_component.use_hl1);
    return btls;
}

/*******************************************************************************************/
void mca_btl_hlink_find_fifo_range (int endpoint_count, int my_rank, int *j_start, int *j_end)
{
    *j_start = *j_end = 0;
return;
    if(endpoint_count == 4)
    {
      if((my_rank % 2) == 0)
      {
        *j_start = 0; *j_end = 1;
      } else {
        *j_start = 2; *j_end = 3;
      }
    } 
}
/*******************************************************************************************/
void mca_btl_hlink_one_fragment_received (hlink_fifo_t *tmp_fifo, mca_btl_hlink_hdr_t *hdr)
{
    mca_btl_hlink_frag_t Frag;
    mca_btl_active_message_callback_t* reg;

    /* recv upcall */
    reg = mca_btl_base_active_message_trigger + hdr->tag;
    Frag.base.des_dst = &Frag.segment;
    Frag.segment.seg_addr.pval = ((char *)hdr) + sizeof(mca_btl_hlink_hdr_t);

    memcpy (&Frag.segment.seg_len, &hdr->len, sizeof(size_t));
    Frag.base.des_dst_cnt = 1;
    reg->cbfunc(&mca_btl_hlink.super, hdr->tag, &(Frag.base), reg->cbdata);
    opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK TIMPIDBG:HLINK component_progress FRAG RCV end");
}
/*******************************************************************************************/
static int mca_btl_hlink_component_progress (void)
{
    mca_btl_hlink_hdr_t *hdr;
    int peer_rank, rc = 0, nevents = 0;
    int j, j_start, j_end;
    int ack_frag = 0;
    char *ptr;
  mca_btl_hlink_find_fifo_range(mca_btl_hlink_component.endpoint_count, mca_btl_hlink_component.my_rank, &j_start, &j_end);

  if(mca_btl_hlink_component.use_hl0)
  {
  //=====================
  // Hyperlink0 interface
  //=====================
    for(j = j_start; j <= j_end; j ++)
    {
        hdr = (mca_btl_hlink_hdr_t *)hlink_fifo_read(mca_btl_hlink_component.fifo[0][j]);
        /* change the address from address relative to the shared
         * memory address, to a true virtual address */
        hdr = RELATIVE2VIRTUAL_0(hdr, j);
        if(hdr == NULL) continue;
        mca_btl_hlink_one_fragment_received (mca_btl_hlink_component.fifo[0][j], hdr);
        nevents ++;
#ifdef BTL_HLINK_VERBOSE
   opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK TIMPIDBG component_progress reading Hyperlink0");
#endif
    }
  }

  if(mca_btl_hlink_component.use_hl1)
  {
    //=====================
    // Hyperlink1 interface
    //=====================
    for(j = j_start; j <= j_end; j ++)
    {
        hdr = (mca_btl_hlink_hdr_t *)hlink_fifo_read(mca_btl_hlink_component.fifo1[0][j]);
        /* change the address from address relative to the shared
         * memory address, to a true virtual address */
        hdr = RELATIVE2VIRTUAL_1(hdr, j);
        if(hdr == NULL) continue;
        mca_btl_hlink_one_fragment_received (mca_btl_hlink_component.fifo1[0][j], hdr);
        nevents ++;
#ifdef BTL_HLINK_VERBOSE
    opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK TIMPIDBG component_progress reading Hyperlink1");
#endif
    }

    if(mca_btl_hlink_component.endpoint_count > 2 && mca_btl_hlink_component.hl_win_msmc_rmt_rd_bridge)
    {
      for(j = j_start; j <= j_end; j ++)
      {
        hdr = (mca_btl_hlink_hdr_t *)hlink_fifo_read(mca_btl_hlink_component.fifo_bridge[0][j]);
        hdr = RELATIVE2VIRTUAL_BRIDGE(hdr, j);
        if(hdr == NULL) continue;
        mca_btl_hlink_one_fragment_received (mca_btl_hlink_component.fifo_bridge[0][j], hdr);
        nevents ++;
#ifdef BTL_HLINK_VERBOSE
      opal_output_verbose(100, mca_btl_base_output, "BTL_HLINK TIMPIDBG component_progress reading diagonal node");
#endif
      }  
    }
  }
  return 1;
}
/*** nothing past this point ***/
