#ifndef _ROUTINGTABLEGEN_H
#define _ROUTINGTABLEGEN_H

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <unistd.h>
#include <stdio.h>
#include <unistd.h>
#include <getopt.h>

#define SRIO_SETUP_SUCCESS  0
#define SRIO_SETUP_FAIL        -1

#define elem_2darray(arr,x,y,cols) (arr[(x) + cols*(y)])

#define MAX_NUM_CARTRIDGES 45
#define MAX_NUM_NODES 4
typedef short Rbool;
#define TRUE 1
#define FALSE 0

/* Generates the id of yth row, xth column */
#define MINX 1

#define srio_node_id_xyz(x,y,z) (MINX + (((y) <<6)+ ((x)<<2) + (z) ))   


#define get_xyz_from_srio_node_id(id,x,y,z)  { z=(id-MINX) &0x3; x=((id-MINX)>>2) &0xf; y=((id-MINX)>>6); }
#define srio_cartridge_id_xy(x,y) ((((y) <<4)+ (x))) 
 


#define MAX_DEPTH 9999

#define BLOCKED MAX_DEPTH

#define ROUTING_SUCCESS 0
#define ROUTING_TABLE_FULL -1
#define ROUTING_TABLE_UNABLE_TO_ADD_AT_INDEX -2
#define ROUTING_TABLE_ERR -1

#define ROUTING_TABLE_INDEX_UNINITIALIZED -1

#define MAX_NUM_CARTRIDGES 45  /* Mooonshot */


#define NUM_MOONSHOT_CARTRIDGES 45
#define NUM_MOONSHOT_ROWS 3
#define NUM_MOONSHOT_COLUMNS 15
#define NUM_MOONSHOT_NODES_WITHIN_CARTRIDGE 4

#define NUM_EVM_CARTRIDGES 1
#define NUM_EVM_ROWS 1
#define NUM_EVM_COLUMNS 1
#define NUM_EVM_NODES_WITHIN_CARTRIDGE 2

/* Walker parameters */
/* Currently initialized as the max nodes in the system. Currently it is moonshot , hence 180 nodes */
#define MAX_WALKER_HOPS_THRESHOLD 180 

/* Routing Table creation Errors */
#define ERR_ROUTING_TABLE_FULL -1
#define ERR_ROUTING_UNREACHABLE_NODE -2

#define SRIO_CLUSTER_FORMATION_SUCCESS 0
#define SRIO_CLUSTER_ERR_RTABLE_ENTRIES_FULL  -1
#define SRIO_CLUSTER_ERR_UNREACHABLE_NODE    -2

typedef enum {
 WALKER_SUCCESS=0,
 WALKER_ERR_FOUND_LOOP=-1,
 WALKER_ERR_CARTRIDGE_DISABLED=-2,
 WALKER_ERR_BLOCKED_PATH=-3,
 WALKER_ERR_UNINITIALIZED_COORDINATES=-4
} srio_walker_retcode_e;

typedef struct {
   int lo;
   int hi;
   int port;
} RoutingTableEntry_t;

typedef struct {
   int num_entries;
   RoutingTableEntry_t entries[8];
} RoutingTable_t;

typedef struct {
   int x;
   int y;
   int z;
} coordinates_t;
/******************************* Connection ***************************/
typedef struct {
  Rbool enabled;
  coordinates_t remoteNode;
  int port_no;
  char node_name[20];
} NodeConnection_t;

typedef struct {
  Rbool enabled;
  coordinates_t remoteCartridge;
  int node_index; /* The node which forms the bridge */
  int port_no;
} CartridgeConnection_t;

/******************************* Node ******************************/

typedef struct {
  Rbool enabled;
  char cartridgeName[20];        /* From json file   Cartridge ID of this node */
  char name[20];                      /* From json file   Node ID */
  int nodeID_Phys;                           /* 1-4 , derived from the CxNx string of the name*/
  char port_connections[4][10]; /* From json file . Port->RemoteNodeName */
  int num_connections;
  NodeConnection_t connections[4]; /* Connections as entered from json file, with coordinates setup */

  Rbool connections_per_dir_initialized; /* Will be set after the Connections assign process is done */
  NodeConnection_t connections_per_dir[4];   /* Populated as a part of Connections assign process  [LEFT/RIGHT] */

  Rbool xyz_assigned; /* Will be assigned after the x-y assign process */
  coordinates_t coordinates;
  int srioID;  
  RoutingTable_t routingTable;
  int outbound_port_dir[4]; /* Outbound port for each direction LEFT/RIGHT/UP/DOWN*/ 
} NodeInfo_t;

/******************************* Cartrdige ***************************/
typedef struct {
  Rbool enabled;  /* Active or not */
  int cartridgeID_XY; /* X-Y ID */
  int cartridgeID_Phys; /* Phys */
  char name[20];    
  int num_nodes;                 /* Number of nodes in this cartrdige, typically 4 */  
  NodeInfo_t nodes[MAX_NUM_NODES]; /* Nodes within this cartrdige */

 #define LEFT    0
  #define RIGHT  1 
  #define UP        2
  #define DOWN  3
  #define SAME 4
  CartridgeConnection_t connections[4]; /* Left/Right/Up/Down cartrdige connections. Will be set after the Connections assign process. */  

  Rbool xy_assigned;   /* Will be assigned after the x-y assign process */
  coordinates_t coordinates;
   RoutingTable_t routingTable;
} Cartridge_t; 

typedef struct {
   int num_rows;
   int num_columns;
   int num_cartridges;
   Cartridge_t cartridges[MAX_NUM_CARTRIDGES];
} Topology_t;

typedef struct {
  char name[10];
} CartridgeXYEntries_t;

typedef struct {
   int num_rows;
   int num_columns;
   CartridgeXYEntries_t *entries;
} CartridgeXY_t;

typedef struct {
    Rbool valid;
    int depth;
} DepthToCartridgeEntries_t;

typedef struct {
  int min_depth_dir;
  DepthToCartridgeEntries_t entries_per_dir[5]; /* LEFT/RIGHT/UP/DOWN/SELF */
} DepthToCartridge_t;

typedef struct {
   int port;
   int depth;
} DepthToDest_t;

typedef struct {
   int dir; /* LEFT/RIGHT/UP/DOWN */
   int depth;
} DepthToNodes_t; /* Depth to nodes within cartridge */

/*********** Structures for Runtime ************/
typedef struct {
   /* The ordered topology, which will be loaded and a cluster generated within */
   Topology_t cluster;
   /* Depth between cartridges */
   DepthToCartridge_t DepthToCartridges[MAX_NUM_CARTRIDGES][MAX_NUM_CARTRIDGES];
   /* Depth between nodes within */
   DepthToNodes_t DepthToNodes[NUM_MOONSHOT_NODES_WITHIN_CARTRIDGE][NUM_MOONSHOT_NODES_WITHIN_CARTRIDGE];
} SRIORunTimeStructures_t;



extern DepthToNodes_t DepthToNodes[NUM_MOONSHOT_NODES_WITHIN_CARTRIDGE][NUM_MOONSHOT_NODES_WITHIN_CARTRIDGE];
extern CartridgeXYEntries_t moonshotXYEntries[NUM_MOONSHOT_ROWS*NUM_MOONSHOT_COLUMNS];
extern CartridgeXY_t moonshotXY;
extern Topology_t  topology_unordered;
extern Topology_t  topology_ordered; /* Ordered as X,Y , Row Major */

int cjson_topology_parse(char *file,Topology_t *topology);
int get_xy_dir(int s_x, int s_y, int d_x, int d_y, int x_max, int y_max);
int cjson_topology_parse(char *file,Topology_t *topology);
int print_topology(Topology_t *topology);
int print_topology_connected(Topology_t *topology);
int topology_assign_xy (Topology_t *topology, CartridgeXY_t *pre_assigned_xy);
int topology_sort (Topology_t *unsorted_topology, Topology_t *sorted_topology);
int topology_establish_connection (Topology_t *topology);

int print_cluster(Topology_t *cluster);
void reset_cluster(Topology_t *cluster);
int find_Path(Topology_t *cluster, Cartridge_t *parent_cartridge,Cartridge_t *my_cartridge, Cartridge_t *dest_cartridge , 
                     int catridge_ids_so_far[],int num_cartridges_so_far, DepthToCartridge_t DepthToCartridge[][MAX_NUM_CARTRIDGES], Rbool *repeat_ret);
void generate_depth_to_cartridges(Topology_t *cluster,DepthToCartridge_t DepthToCartridges[][MAX_NUM_CARTRIDGES]);
void print_depth_to_cartridges  (Topology_t *cluster,DepthToCartridge_t depthToCartridges[][MAX_NUM_CARTRIDGES],
                                               int num_cartridges);
void init_depth_to_cartridges  (DepthToCartridge_t depthToCartridges[MAX_NUM_CARTRIDGES][MAX_NUM_CARTRIDGES],
                                               int num_cartridges);
void init_routing_table  (RoutingTable_t *routingTable);
void enable_cartridges_by_name(Topology_t *topology, char active_cartridges[][10], int num_active_cartridges);
void enable_cartridges_by_PhysID(Topology_t *topology, int active_cartridges[], int num_active_cartridges);
void map_physID_to_SRIOID(Topology_t *topology, int active_cartridges[], int num_active_cartridges);
int add_to_routing_table_at_index(RoutingTable_t *routingTable,int dest_srioID,int routing_port,int last_routing_table_index);
int add_new_to_routing_table(RoutingTable_t *routingTable,int dest_srioID,int port);
void print_routing_table(NodeInfo_t *this_node, RoutingTable_t *routingTable);
int find_routing_table_index(Cartridge_t *this_cartridge,
                                             Cartridge_t *dest_cartridge, 
                                             NodeInfo_t *this_node, 
                                             int dest_srioID,
                                             int last_routing_table_index,
                                             DepthToCartridge_t *depthToDestCartridge, 
                                             DepthToNodes_t DepthToNodes[4][4],
                                             int *routing_table_index);
int generate_routing_entries (Topology_t *cluster, 
                                                DepthToCartridge_t depthToCartridges[MAX_NUM_CARTRIDGES][MAX_NUM_CARTRIDGES],
                                                int num_cartridges);

void * srioUtils_LoadFromFile(FILE *topology_file);
int  srioUtils_map_physID_to_SRIOId(void *base,int cartridge,int node,unsigned short *srioID);
int  srioUtils_map_SRIOId_to_physID(void *base,int srioID,int *cartridge,int *node);
int  srioUtils_find_routing_port(const void *base,int srio_id_src, int srio_id_dest);
int srioUtils_formCluster(void *base,int cartridge_phys_ids[],int num_cartridges);

int srioUtils_displayPortForwardingRegs(void *base, void *srio_base_virtual, int mySrioID);
int  srioUtils_progFwdingRegs(void *base,  void *srio_base_virtual, int mySrioID);
int  srioUtils_progFwdingRegsCustom(void *base,  void *srio_base_virtual, int mySrioID,RoutingTable_t *routingTable);
int  srioUtils_resetFwdingRegs(void *base,  void *srio_base_virtual, int mySrioID);
int srioUtils_displayCluster(void *base);
int srioUtils_ResetCluster(void *base);


/************************************************************************************************/
/*                       Utility Functions                                                                                                                  */
/************************************************************************************************/
static inline Rbool cartridge_present(int cartridgeID_XY,int cartridge_ids_so_far[],int num_cartridges_so_far) {
  int i;
  Rbool found=FALSE;
  for(i=0;i<num_cartridges_so_far;i++) {
      if(cartridgeID_XY==cartridge_ids_so_far[i]) {
         found=TRUE;
         break;
      }
  }
  return found;
}

/************************************************************************************************/
static inline Rbool find_node_by_name(Topology_t *topology,char name[],coordinates_t *coordinates_ret) {
   
     int x,y,z,cartridge_no,node_no;
     Rbool found;
     coordinates_t coordinates ={-1,-1,-1};
    Cartridge_t * this_cartridge;
    NodeInfo_t *this_node;
     found=FALSE;
      for(cartridge_no=0;cartridge_no< topology->num_cartridges; cartridge_no++) {
           /* Display Cartridge Info */
           this_cartridge=&(topology->cartridges[cartridge_no]);
           
         for(node_no=0;node_no< this_cartridge->num_nodes; node_no++) {
             this_node=&this_cartridge->nodes[node_no];
             if(!strcmp(this_node->name,name)) {
                             found=TRUE;
                             coordinates=this_node->coordinates; 
                             break;
               }

           }
              if(found) 
                  break;
       }

       *coordinates_ret=coordinates;
       if(!found) {
          printf("\n Name  Not found for %s",name);
       }
       return(found);
}
/************************************************************************************************/
static inline Cartridge_t * find_Cartridge_by_coordinates(Topology_t *topology,coordinates_t *coordinates) {
     int x,y,z,cartridge_no;
     Cartridge_t *cartridge_elem;
  
            cartridge_elem=&(elem_2darray(topology->cartridges,
                                                               coordinates->x,
                                                               coordinates->y,
                                                               topology->num_columns));
             return(cartridge_elem);
    
}


/************************************************************************************************/
static inline NodeInfo_t * find_Node_by_coordinates(Topology_t *topology,coordinates_t *coordinates) {
     int x,y,z,cartridge_no;
     Cartridge_t *cartridge_elem;
  
            cartridge_elem=&(elem_2darray(topology->cartridges,
                                                               coordinates->x,
                                                               coordinates->y,
                                                               topology->num_columns));
             return(&(cartridge_elem->nodes[coordinates->z]));
    
}



#endif





