#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <unistd.h>
#include <stdio.h>
#include <unistd.h>
#include <getopt.h>
#include <sys/mman.h>
#include <errno.h>

#include "routingTableGen.h"
#include "mpi_interfaces.h"

#ifndef TRUE
  #define TRUE 1
#endif 
 
#ifndef FALSE
   #define FALSE 0
#endif

/****************************************************************************************************
*******************************  Map SRIO Regs  *************************************************
****************************************************************************************************/
uint32_t srio_mmap_uio(uint32_t addr, uint32_t size,int uio_mem_fd)
{
        uint32_t virt_addr;
        uint32_t page_size;
        page_size = sysconf(_SC_PAGE_SIZE);
        if (size%page_size)
        {
                printf("Size does not align with page size. Size given: %d\n", size);
                return 0;
        }
        if ((uint32_t)addr % page_size)
        {
                printf("Address does not align with page size. Address given: 0x%08x\n", (uint32_t) addr);
                return 0;
        }
        virt_addr = (uint32_t) mmap(0, size, (PROT_READ|PROT_WRITE), MAP_SHARED, uio_mem_fd, (off_t)addr);
        if (virt_addr == -1)
        {
                printf("mmap failed!\n");
                return 0;
        }
        return virt_addr;
}

/****************************************************************************************************
*******************************  Map SRIO Regs  *************************************************
****************************************************************************************************/
int main(int argc, char *argv[])
{
	int cmd;
	int fd, res;
	char *routing_config_file_str;
        int i,ret;
        FILE *routing_config_file=NULL;
       void *base=NULL;
       int cartridge_id  /*IN*/, node_id /*IN*/;
      unsigned short srioID,srioID1,srioID2,port;
      int myID;
      uint32_t srio_base_real=0x2900000;
      uint32_t *srio_base_virtual;
      int uio_mem_fd;
      Rbool programRoutingTables=TRUE, resetRoutingTables=FALSE;


      FILE *routerCfgFile=NULL;
     /* Find myID */

        if (argc <=1) {
 		perror("Insufficient arguments: Usage:  pktFwdUtil <display/reset/routingfile/custom> <myID> <routingConfigfile> ");	
                 exit(-2);
          }

     uio_mem_fd = open("/dev/srio", (O_RDWR | O_SYNC));
     if (fd < 0) {
	printf("\n UIO: Error opening device ");
	return;
     }

     srio_base_virtual = (uint32_t *)srio_mmap_uio((uint32_t)srio_base_real, 0x1c000,uio_mem_fd);
     printf("\n UIO Memmapped Real = %x,  Virtual = %p",srio_base_real,srio_base_virtual);
     if (srio_base_virtual == MAP_FAILED) {
	printf("mmap failed");
     }

     if(strcmp(argv[1],"reset")==0) {
  
        printf("\n  ** Reset Port Forwarding registers ***\n");
        SrioResetPktFwdingRegs(base, srio_base_virtual,myID);
        SrioDisplayPktFwdingRegs(base,srio_base_virtual,myID);

     } else if(strcmp(argv[1],"routingfile")==0) {
        int cartridge_phys_ids[]={1,4,7};
       int num_cartridge_phys_ids=sizeof(cartridge_phys_ids)/sizeof(int);

       printf("\n  ** Program with RoutingTable from file ***\n");
       sscanf(argv[2], "%d", &myID);
       printf("\n My NodeID = %d",myID);
       routing_config_file_str = argv[3];

        routing_config_file = fopen(routing_config_file_str,"rb");
        printf("\nCreating Routing structures");
        base = SrioRoutingStructuresCreate(routing_config_file_str);
        printf("\n Forming Cluster ");
       ret=SrioCreateRoutingTableFromCluster(base,cartridge_phys_ids,num_cartridge_phys_ids);

#if 0
        srioUtils_display_routingTable(&(((srioConfigDataStructures_t *)base)->routingTableHeader));
#endif
        SrioDisplayPktFwdingRegs(base,srio_base_virtual,myID);
         SrioProgramPktFwdingRegs(base, srio_base_virtual,myID);
        SrioDisplayPktFwdingRegs(base,srio_base_virtual,myID);

     } else if(strcmp(argv[1],"display")==0) {

        printf("\n  ** Display Port Forwarding registers ***\n");
        SrioDisplayPktFwdingRegs(base,srio_base_virtual,myID);

     } else if(strcmp(argv[1],"custom")==0) {
        RoutingTable_t routingTablePerNode;

        SrioResetPktFwdingRegs(base, srio_base_virtual,myID);
         routingTablePerNode.num_entries=0;

        int num_vals=argc-2;
        int arg_index=2;
       while(num_vals > 0) {
            int lo,hi,port;
            sscanf(argv[arg_index++], "%d", &lo);
            sscanf(argv[arg_index++], "%d", &hi);
            sscanf(argv[arg_index++], "%d", &port);

           routingTablePerNode.entries[routingTablePerNode.num_entries].lo=lo;
           routingTablePerNode.entries[routingTablePerNode.num_entries].hi=hi;
           routingTablePerNode.entries[routingTablePerNode.num_entries].port=port;
           routingTablePerNode.num_entries++;

           num_vals= num_vals-3;
        }
      printf("\n Found %d entries to program the routing registers ..",routingTablePerNode.num_entries);

/*        routingTablePerNode.entries[routingTablePerNode.num_entries].lo=2;
        routingTablePerNode.entries[routingTablePerNode.num_entries].hi=2;
        routingTablePerNode.entries[routingTablePerNode.num_entries].forwardingPort=1;
        routingTablePerNode.num_entries++;

        routingTablePerNode.entries[routingTablePerNode.num_entries].lo=0;
        routingTablePerNode.entries[routingTablePerNode.num_entries].hi=0;
        routingTablePerNode.entries[routingTablePerNode.num_entries].forwardingPort=0;
        routingTablePerNode.num_entries++;
*/
     


        srioUtils_progFwdingRegsCustom(base, srio_base_virtual,myID,&routingTablePerNode);
        SrioDisplayPktFwdingRegs(base,srio_base_virtual,myID);
     }  

     printf("\n ");
     exit(0);
}




